#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "types.h"
#include "proto.h"
#include "select.h"
#include "matrix.h"
#include "f.h"
#include "transform.h"
#include "print.h"
#include "mean-groups.h"
#include "variance.h"
#include "statistics.h"

/* local prototypes */
static real sum_of_all(Matrix * data[], const matrix_dim max_depth);
static bool F_is_significant(const matrix_dim dof_num, 
              const matrix_dim dof_den, const real value, real * critval,
              const bool showall);
static real SSx(real value[], const matrix_dim howmany, 
              const matrix_dim denom);
static real SSxy(real * value[], const matrix_dim first, 
              const matrix_dim second, const matrix_dim denom);

/* local macros */
#define Square(X) ((X) * (X))

/* module specific global variables */
static const double alpha_level = 0.05;


/*                                                                         *\
   -----------------------------------------------------------------------
                                  F-test
   -----------------------------------------------------------------------
\*                                                                         */

bool F_is_significant(const matrix_dim dof_num, /* deg.freedom in numerator */
                      const matrix_dim dof_den, /* d.o.f. in denom. */
                      const real value,         /* critical value to compare  
                                                  against table */
                            real *Fval,        /* critical value in table */
                      const bool showall)
{
   const  char proc[] ="F_is_significant";
   extern char errorstring[];

   if ((dof_num > INT_MAX) || (dof_den > INT_MAX)) {
      sprintf(errorstring, "cannot computer an F value for v1="MAT_DIM_SPEC \
                           " and v2="MAT_DIM_SPEC" -- there would be "      \
                           "numeric overflow.",                             \
                           dof_den, dof_num);
      swerr(proc, errorstring, 1);
      /*NOTREACHED*/
   }

   *Fval = (real)critf(alpha_level, (int)dof_num, (int)dof_den);
   if (showall) {
      printf("   F[v1="MAT_DIM_SPEC";v2="MAT_DIM_SPEC";alpha=%5.3f]" \
             " = "POS_REAL_SPEC,                                     \
             dof_num, dof_den, alpha_level, *Fval);
      NewLine(2, stdout);
   }
   return (value > *Fval);
} /* F_is_significant() */


/*                                                                         *\
   -----------------------------------------------------------------------
                               Two Way ANOVA
   -----------------------------------------------------------------------
\*                                                                         */


real SSx(      real       value[],
         const matrix_dim howmany,
         const matrix_dim denom)
{
   real         sum = 0.0;
   unsigned int i;

   for (i=0; i < howmany; i++) {
      sum += Square(value[i]);
   }
   sum /= (real)denom;
   return(sum);
} /* SSx() */


void TwoWayStats(      Matrix * mat,
                 const string   datatype, /* "Average" or "R-" */
                 const string   partiname[],
                 const string   queryname[],
                 const string   levelname,
                       char     trans_str[],
                 const bool     showall)
{
   const    char        proc[] = "TwoWayStats";
            ATableEntry info[4];
            real        partisum[MAX_PARTICIPANT],
                        partimean[MAX_PARTICIPANT];
            real        querysum[NUMofQUERIES],
                        querymean[NUMofQUERIES];
            real        xdotdot = 0.0;
            real        Q;
   register matrix_dim  parti;
   register matrix_dim  query;
            matrix_dim  maxparti = num_cols(mat);
            matrix_dim  maxquery = num_rows(mat);
            real        partivar[MAX_PARTICIPANT];
            Means       averages[MAX_PARTICIPANT];
            char        transstring[MAXscreenwidth + 1];
            char        levelstr[MAXscreenwidth + 1];
            bool        var_test_okay,
                        quit = FALSE;
            enum {PARTI = 0U, QUERY = 1U, WITHIN = 2U, TOTAL = 3U};

   strncpy(transstring, "Raw Data", MAXscreenwidth);
   transstring[MAXscreenwidth] = '\0';
   end_display();
   if (showall) {
      fputs("  ===================", stdout);
      NewLine(1, stdout);
      fputs("    Two Way ANOVA    ", stdout);
      NewLine(1, stdout);
      fputs("  ===================", stdout);
      NewLine(1, stdout);
      printf("== %s", datatype);
      if (levelname) {
         printf(" (%s)", levelname);
      }
      printf(" ==");
      NewLine(2, stdout);
   }

   info[PARTI].srcvar    = "Participants";
   info[PARTI].degfreed  = maxparti - 1;
   info[QUERY].srcvar    = "Queries";
   info[QUERY].degfreed  = maxquery - 1;
   info[WITHIN].srcvar   = "Within Cells";
   info[WITHIN].degfreed = info[PARTI].degfreed * info[QUERY].degfreed;
   info[TOTAL].srcvar    = "Total";
   info[TOTAL].degfreed  = (maxparti * maxquery) - 1;

   for (parti=0; parti < maxparti; parti++) {
      partisum[parti] = sumcol(mat, parti);
   }
   for (parti=0; parti < maxparti; parti++) {
      partimean[parti] = partisum[parti] / maxquery;
   }
   variance(partisum, mat, maxparti, maxquery, TRUE,  partivar);
   /* variance(querysum, mat, maxquery, maxparti, FALSE, queryvar); */

   if ((!variances_are_equal(&mat, 1, partimean, partivar, showall, \
           &var_test_okay)) && (var_test_okay)) {
      printf("Variances of participant data are not equal.");
      NewLine(1, stdout);
      if (showall) {
         if (NULL != transform(TRANS_TEST)) {
            printf("Applying %s transform.", tran_name());
	 } else {
            fputs("Not applying any transform.", stdout);
	 }
         NewLine(1, stdout);
      }
      if (NULL != transform(TRANS_TEST)) {
         mat_forall(&mat, transform(TRANS_TEST), 1);
         sprintf(transstring, "After %s transform applied", tran_name());
         strncpy(transstring, (trans_str)?trans_str:"", MAXscreenwidth);
         transstring[MAXscreenwidth] = '\0';
      }

      for (parti=0; parti < maxparti; parti++) {
         partisum[parti] = sumcol(mat, parti);
      }
      for (parti=0; parti < maxparti; parti++) {
         partimean[parti] = partisum[parti] / maxquery;
      }
   } else {
      if (! var_test_okay) {
         swerr(proc, \
               "can't test equality of variances -- assuming equality", 0);
      }
   }

   for (query=0; query < maxquery; query++) {
      querysum[query] = sumrow(mat, query);
   }
   for (query=0; query < maxquery; query++) {
      querymean[query] = querysum[query] / maxparti;
   }

   info[PARTI].sumsquare = SSx(partisum, maxparti, maxquery);
   info[QUERY].sumsquare = SSx(querysum, maxquery, maxparti);
   info[WITHIN].sumsquare = 0.0;
   for (parti=0; parti < maxparti; parti++) {
      for (query=0; query < maxquery; query++) {
         real value = mat_value(mat, query, parti);
         info[WITHIN].sumsquare += Square(value);
      }
   }
   xdotdot = sum_of_all(&mat, 1);   

   if (0 == info[PARTI].degfreed) {
      fputs("Zero degrees of freedom for participants -- can't continue.", \
            stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }
   if (0 == info[QUERY].degfreed) {
      fputs("Zero degress of freedom for queries -- can't continue.", \
            stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }

   xdotdot = Square(xdotdot) / (maxparti * maxquery);
   info[TOTAL].sumsquare = info[WITHIN].sumsquare - xdotdot;
   info[WITHIN].sumsquare -= (info[PARTI].sumsquare + info[QUERY].sumsquare);
   info[WITHIN].sumsquare += xdotdot;

   if (0 == info[WITHIN].sumsquare) {
      fputs("Zero SSW -- can't continue.", stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }
   if (quit) {
      end_display();
      return;
   }

   info[WITHIN].meansquare = info[WITHIN].sumsquare / \
                             (info[PARTI].degfreed * info[QUERY].degfreed);
   info[PARTI].sumsquare -= xdotdot;
   info[QUERY].sumsquare -= xdotdot;
   info[PARTI].meansquare = info[PARTI].sumsquare / info[PARTI].degfreed;
   info[QUERY].meansquare = info[QUERY].sumsquare / info[QUERY].degfreed;

   /* participant means */
   info[PARTI].f_value = info[PARTI].meansquare / info[WITHIN].meansquare;
   if (showall) {
      NewLine(1, stdout);
      fputs(" ------------------------------", stdout);
      NewLine(1, stdout);
      fputs("   F-test of participant data  ", stdout);
      NewLine(1, stdout);
      fputs(" ------------------------------", stdout);
      NewLine(1, stdout);
      printf("   F = MSA/MSW = "REAL_SPEC, info[PARTI].f_value);
      NewLine(1, stdout);
   }

   info[PARTI].signifdiff = F_is_significant(info[PARTI].degfreed,   \
                                             info[WITHIN].degfreed,  \
                                             info[PARTI].f_value,    \
                                             &(info[PARTI].critval), \
                                             showall);

   /* query means */
   info[QUERY].f_value = info[QUERY].meansquare / info[WITHIN].meansquare;
   if (showall) {
      NewLine(1, stdout);
      fputs(" -----------------------", stdout);
      NewLine(1, stdout);
      fputs("   F-test of query data ", stdout);
      NewLine(1, stdout);
      fputs(" -----------------------", stdout);
      NewLine(1, stdout);
      printf("   F = MSB/MSW = "REAL_SPEC, info[QUERY].f_value);
      NewLine(1, stdout);
   }
   info[QUERY].signifdiff = F_is_significant(info[QUERY].degfreed, \
                                             info[WITHIN].degfreed,\
                                             info[QUERY].f_value,  \
                                             &info[QUERY].critval, \
                                             showall);
   if (levelname) {
      sprintf(levelstr, "%s, level = %s", datatype, levelname);
   } else {
      sprintf(levelstr, "%s", datatype);
   }

   /* PARTI groups */
   Q = HSD(info[PARTI].signifdiff, info[PARTI].degfreed, info[QUERY].degfreed,\
           info[WITHIN].meansquare, alpha_level, showall);
   if (make_means(averages, partiname, partimean, maxparti)) {
      mean_groups(Q, averages, maxparti);
      show_means(averages, maxparti, levelstr,        \
                 (transstring[0])?transstring:NULL,   \
                 "participant", NULL, showall, stdout);
   } else {
      swerr(proc, "unable to make mean groups", 0);
   }

   /* QUERY groups */
   if (showall) {
      if (info[QUERY].signifdiff) {
         fputs("NOTE: At least one pair of the query means differ.", stdout);
      } else {
         fputs("No significant difference detected between query means.", \
               stdout);
      }
      NewLine(1, stdout);
   }

   ANOVATable(info, (TOTAL + 1U),                             \
              "ANOVA Table for a Difference in Mean Scores, " \
              "Two Factor Design", levelstr, NULL,            \
              (transstring[0])?transstring:NULL, stdout);                                   

   if (showall) { 
      if (1 < maxparti) {
         print_name_list(partiname, maxparti, 5, width_display(), \
                         "Participants: ", stdout);
      }
      if (1 < maxquery) {
         print_name_list(queryname, maxquery, 5, width_display(), \
                         "Queries: ", stdout);
      }
   }
   end_display();
}/* TwoWayStats() */

/*                                                                         *\
   -----------------------------------------------------------------------
                              Three Way ANOVA
   -----------------------------------------------------------------------
\*                                                                         */

real SSxy(      real *     value[], 
          const matrix_dim first,
          const matrix_dim second,
          const matrix_dim denom)
{
    real         sum = 0.0;
    unsigned int i, j;

   for (i=0; i < first; i++) {
      for (j=0; j < second; j++) {
         sum += Square(value[i][j]);
      }
   }
   sum /= (real)denom;
   return(sum);
} /* SSxy() */


void ThreeWayStats(      Matrix * mat[],
                 const matrix_dim maxlevel, /* depth of matrix */
                 const string     datatype,
                 const string     level_list,
                 const string     partiname[],
                 const string     queryname[],
                       char       trans_str[],
                 const bool       showall)
{
            enum {PARTI  = 0U, QUERY  = 1U, RECALL = 2U, \
                  PARQUE = 3U, PARREC = 4U, QUEREC = 5U, \
                  WITHIN = 6U, TOTAL = 7};
   const    char          proc[] = "ThreeWayStats";
   const    bool          always_show_means = TRUE;
            ATableEntry   info[8];
            unsigned long denom[6];
            real          partisum[MAX_PARTICIPANT],
                          partimean[MAX_PARTICIPANT];
            real          querysum[NUMofQUERIES],
                          querymean[NUMofQUERIES];
            real          recallsum[NUMofQUERIES],
                          recallmean[NUMofQUERIES];
            real *        parquesum[MAX_PARTICIPANT];
            real *        parquemean[MAX_PARTICIPANT];
            real *        parrecsum[MAX_PARTICIPANT];
            real *        parrecmean[MAX_PARTICIPANT];
            real *        querecsum[NUMofQUERIES];
            real *        querecmean[NUMofQUERIES];
            real          xdotdotdot = 0.0;
            real          Q;
   register matrix_dim    parti;
   register matrix_dim    query;
   register matrix_dim    level; /* level of recall parameter */
            matrix_dim    maxparti = num_cols(mat[0]);
            matrix_dim    maxquery = num_rows(mat[0]);
            real          parquevar[MAX_PARTICIPANT][NUMofQUERIES],
                          variance[MAX_PARTICIPANT * NUMofQUERIES];
            Means         averages[MAX_PARTICIPANT];
            char          transstring[MAXscreenwidth + 1];
            char          parti_descrip[MAXscreenwidth];
            char          query_descrip[MAXscreenwidth];
            bool          var_test_okay,
                          quit = FALSE,
                          no_signif_diff = TRUE;
            
   strncpy(transstring, "Raw Data", MAXscreenwidth);
   transstring[MAXscreenwidth] = '\0';
   info[PARREC].srcvar = parti_descrip;
   info[QUEREC].srcvar = query_descrip;
   end_display();
   if (showall) {
      fputs("  =====================", stdout);
      NewLine(1, stdout);
      fputs("     Three Way ANOVA   ", stdout);
      NewLine(1, stdout);
      fputs("  =====================", stdout);
      NewLine(1, stdout);
      printf("  == %s ==", datatype);
      NewLine(1, stdout);
      printf("   ==  %s  ==", level_list);
      NewLine(2, stdout);
   }

   {
      register matrix_dim p, q;

      for (p=0; p < maxparti; p++) {
         if ((NULL == (parquesum[p]  = \
                      (real *)calloc(NUMofQUERIES, sizeof(real)))) || \
             (NULL == (parquemean[p] = \
                      (real *)calloc(NUMofQUERIES, sizeof(real)))) || \
             (NULL == (parrecsum[p]  = \
                      (real *)calloc(MAX_LEVEL,    sizeof(real)))) || \
             (NULL == (parrecmean[p] = \
                      (real *)calloc(MAX_LEVEL,    sizeof(real)))) )  {
            swerr(proc, "Not enough memory for sum and mean arrays!", 0);
            return;
         }
      }
      for (q=0; q < maxquery; q++) {
         if ( (NULL == (querecsum[q]  = \
                       (real *)calloc(MAX_LEVEL, sizeof(real)))) || \
              (NULL == (querecmean[q] = \
                       (real *)calloc(MAX_LEVEL, sizeof(real)))))   {
            swerr(proc, "Not enough memory for sum and mean arrays!", 0);
            return;
	 }
      }
   }
   denom[PARTI]  = (maxquery * maxlevel);
   denom[QUERY]  = (maxparti * maxlevel);
   denom[RECALL] = (maxparti * maxquery);
   denom[PARQUE] = maxlevel;
   denom[PARREC] = maxquery;
   denom[QUEREC] = maxparti;

   info[PARTI].srcvar    = "Participants";
   info[PARTI].degfreed  = maxparti - 1;
   info[PARTI].smallname = "participant";
   info[QUERY].srcvar    = "Queries";
   info[QUERY].degfreed  = maxquery - 1;
   info[QUERY].smallname = "query";
   info[RECALL].degfreed = maxlevel - 1;
   info[RECALL].srcvar   = "Measure";
   info[RECALL].smallname= "recall";
   info[PARQUE].degfreed = info[PARTI].degfreed * info[QUERY].degfreed;
   info[PARQUE].srcvar   = "Interaction of Participants & Queries";
   info[PARREC].srcvar   = "Interaction of Participants & Measure";
   info[PARREC].degfreed = info[PARTI].degfreed * info[RECALL].degfreed;
   info[QUEREC].srcvar   = "Interaction of Queries & Measure";
   info[QUEREC].degfreed = info[QUERY].degfreed * info[RECALL].degfreed;
   info[WITHIN].srcvar   = "Within Cells";
   info[WITHIN].degfreed = info[PARTI].degfreed * info[QUERY].degfreed \
                           * info[RECALL].degfreed;
   info[TOTAL].srcvar    = "Total";
   info[TOTAL].degfreed  = (maxparti * maxquery * maxlevel) - 1;

   if (0 == info[PARTI].degfreed) {
      fputs("Zero degress of freedom for participants -- can't continue.", \
            stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }
   if (0 == info[QUERY].degfreed) {
      fputs("Zero degress of freedom for queries -- can't continue.", \
            stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }
   if (0 == info[RECALL].degfreed) {
      fputs("Zero degress of freedom for recall level -- can't continue.", \
            stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }

   for (parti=0; parti < maxparti; parti++) {
      partisum[parti] = sum3Dcol(mat, parti, maxlevel);
   }
   for (parti=0; parti < maxparti; parti++) {
      partimean[parti] = partisum[parti] / denom[PARTI];
   }

   /* compute sum_of_squares for use in ThreeWay_variance() */
   /*
      initial value of info[W].ss is 
         Sum h=1..parti Sum i..query Sum j..level X^2_hij
   */
   info[WITHIN].sumsquare = 0.0;
   for (parti=0; parti < maxparti; parti++) {
      for (query=0; query < maxquery; query++) {
         for (level=0; level < maxlevel; level++) {
            real value = mat_value(mat[level], query, parti); 

            info[WITHIN].sumsquare += Square(value);
	 }
      }
   }

   ThreeWay_variance(info[WITHIN].sumsquare, mat, \
                     maxparti, maxquery, maxlevel, parquevar);

   for (parti=0; parti < maxparti; parti++) {
      for (query=0; query < maxquery; query++) {
         variance[(parti * query) + query] = parquevar[parti][query];
      }
   }
   if ((!variances_are_equal(mat, maxlevel, partimean, variance, showall, \
        &var_test_okay)) && (var_test_okay)) {
      printf("Variances of participant data are not equal.");
      NewLine(1, stdout);
      if (showall) {
         if (NULL != transform(TRANS_TEST)) {
            printf("Applying %s transform.", tran_name());
	 } else {
            printf("%s", "Not applying any transform.");
	 }
         NewLine(2, stdout);
      }
      if (NULL != transform(TRANS_TEST)) {
         mat_forall(mat, transform(TRANS_TEST), maxlevel);
         sprintf(transstring, "After %s transform applied", tran_name());
         strncpy(transstring, (trans_str)?trans_str:"", MAXscreenwidth);
         transstring[MAXscreenwidth] = '\0';

         /* the data has changed, so the SS within needs to be recomputed */
         info[WITHIN].sumsquare = 0.0;
         for (parti=0; parti < maxparti; parti++) {
            for (query=0; query < maxquery; query++) {
               for (level=0; level < maxlevel; level++) {
                  real value = mat_value(mat[level], query, parti); 

                  info[WITHIN].sumsquare += Square(value);
	       }
            }
         }
      }
      for (parti=0; parti < maxparti; parti++) {
         partisum[parti] = sum3Dcol(mat, parti, maxlevel);
      }
      for (parti=0; parti < maxparti; parti++) {
         partimean[parti] = partisum[parti] / denom[PARTI];
      }
   } else {
      if (! var_test_okay) {
         swerr(proc, 
               "can't test equality of variances -- assuming equality", 0);
       }
   }

   if (0 == info[WITHIN].sumsquare) {
      fputs("Zero SSW -- can't continue.", stdout);
      NewLine(1, stdout);
      quit = TRUE;
   }
   if (quit) {
      end_display();
      return;
   }

   /* Second Factor */
   for (query=0; query < maxquery; query++) {
      querysum[query] = sum3Drow(mat, query, level);
   }
   for (query=0; query < maxquery; query++) {
      querymean[query] = querysum[query] / denom[QUERY];
   }

   /* Third Factor */
   for (level=0; level < maxlevel; level++) {
      recallsum[level] = sum_of_all(&(mat[level]), 1);
   }
   for (level=0; level < maxlevel; level++) {
      recallmean[level] = recallsum[level] / denom[RECALL];
   }

   /* Interaction of first and second factors */
   for (parti=0; parti < maxparti; parti++) {
      for (query=0; query < maxquery; query++) {
         parquesum[parti][query] = sumdepth(mat, query, parti, maxlevel);
      }
   }
   for (parti=0; parti < maxparti; parti++) {
      for (query=0; query < maxquery; query++) {
         parquemean[parti][query] = parquesum[parti][query] / denom[PARQUE];
      }
   }

   /* Interaction of first and third factors */
   for (parti=0; parti < maxparti; parti++) {
      for (level=0; level < maxlevel; level++) {
         parrecsum[parti][level] = sumcol(mat[level], parti);
      }
   }
   for (parti=0; parti < maxparti; parti++) {
      for (level=0; level < maxlevel; level++) {
         parrecmean[parti][level] = parrecsum[parti][level] / denom[PARREC];
      }
   }

   /* Interaction of second and third factors */
   for (query=0; query < maxquery; query++) {
      for (level=0; level < maxlevel; level++) {
         querecsum[query][level] = sumrow(mat[level], query);
      }
   }
   for (query=0; query < maxquery; query++) {
      for (level=0; level < maxlevel; level++) {
         querecmean[query][level] = querecsum[query][level] / \
                                       (real)denom[QUEREC];
      }
   }

   info[PARTI].sumsquare  = SSx(partisum,  maxparti, denom[PARTI]);
   info[QUERY].sumsquare  = SSx(querysum,  maxquery, denom[QUERY]);
   info[RECALL].sumsquare = SSx(recallsum, maxlevel, denom[RECALL]);
   info[PARQUE].sumsquare = SSxy(parquesum, maxparti, maxquery, denom[PARQUE]);
   info[PARREC].sumsquare = SSxy(parrecsum, maxparti, maxlevel, denom[PARREC]);
   info[QUEREC].sumsquare = SSxy(querecsum, maxquery, maxlevel, denom[QUEREC]);
   xdotdotdot = sum_of_all(mat, maxlevel);

   xdotdotdot = Square(xdotdotdot) / (maxparti * maxquery * maxlevel);
   info[TOTAL].sumsquare = info[WITHIN].sumsquare - xdotdotdot;
   info[WITHIN].sumsquare -= info[PARQUE].sumsquare;
   info[WITHIN].sumsquare -= info[PARREC].sumsquare;
   info[WITHIN].sumsquare -= info[QUEREC].sumsquare;
   info[WITHIN].sumsquare += info[PARTI].sumsquare;
   info[WITHIN].sumsquare += info[QUERY].sumsquare;
   info[WITHIN].sumsquare += info[RECALL].sumsquare;
   info[WITHIN].sumsquare -= xdotdotdot;
   info[WITHIN].meansquare = info[WITHIN].sumsquare / \
                             (info[PARTI].degfreed    \
                             * info[QUERY].degfreed   \
                             * info[RECALL].degfreed);

   info[PARQUE].sumsquare -= (info[PARTI].sumsquare + info[QUERY].sumsquare);
   info[PARQUE].sumsquare += xdotdotdot;
   info[PARQUE].meansquare = info[PARQUE].sumsquare / info[PARQUE].degfreed;

   info[PARREC].sumsquare -= (info[PARTI].sumsquare + info[RECALL].sumsquare);
   info[PARREC].sumsquare += xdotdotdot;
   info[PARREC].meansquare = info[PARREC].sumsquare / info[PARREC].degfreed;

   info[QUEREC].sumsquare -= (info[QUERY].sumsquare + info[RECALL].sumsquare);
   info[QUEREC].sumsquare += xdotdotdot;
   info[QUEREC].meansquare = info[QUEREC].sumsquare / info[QUEREC].degfreed;


   info[PARTI].sumsquare  -= xdotdotdot;
   info[PARTI].meansquare  = info[PARTI].sumsquare / info[PARTI].degfreed;
   info[QUERY].sumsquare  -= xdotdotdot;
   info[QUERY].meansquare  = info[QUERY].sumsquare / info[QUERY].degfreed;
   info[RECALL].sumsquare -= xdotdotdot;
   info[RECALL].meansquare = info[RECALL].sumsquare / info[RECALL].degfreed;


  /***
   F-tests 
     compute f_value, signifdiff and critval
  ****/

#define simple_F_test(FACTOR) {                                              \
   if (showall) {                                                            \
      const char   before[] = "   F-test of ";                               \
      const char   after[]  = " data ";                                      \
      const size_t baselen  = strlen(before) + strlen(after);                \
                                                                             \
      fputs(" ", stdout);                                                    \
      PrintLine(stdout, '-', strlen(info[FACTOR].smallname)+baselen, "");    \
      NewLine(1, stdout);                                                    \
      printf("%s%s%s", before, info[FACTOR].smallname, after);               \
      NewLine(1, stdout);                                                    \
      fputs(" ", stdout);                                                    \
      PrintLine(stdout, '-', strlen(info[FACTOR].smallname)+baselen, "");    \
      NewLine(1, stdout);                                                    \
   }                                                                         \
   info[FACTOR].f_value = info[FACTOR].meansquare / info[WITHIN].meansquare; \
   if (showall) {                                                            \
      printf("   F = MS%c/MSW = "REAL_SPEC,                                  \
             ('A'+FACTOR), info[FACTOR].f_value);                            \
      NewLine(1, stdout);                                                    \
   }                                                                         \
   info[FACTOR].signifdiff = F_is_significant(info[FACTOR].degfreed,         \
                                             info[WITHIN].degfreed,          \
                                             info[FACTOR].f_value,           \
                                             &(info[FACTOR].critval),        \
                                             showall);                       \
   if (showall) {                                                            \
      if (info[FACTOR].signifdiff) {                                         \
         printf("    At least one pair of the %s means differ.",             \
                info[FACTOR].smallname);                                     \
      } else {                                                               \
         printf("    No significant difference detected between %s "         \
                "means.",                                                    \
                info[FACTOR].smallname);                                     \
      }                                                                      \
      NewLine(2, stdout);                                                    \
   }                                                                         \
} /* #simple_F_test */

   /* If F=MSA/MSW > F[(a-1), (a-1)(b-1)(c-1); alpha] reject H_01 */
   simple_F_test(PARTI);
   /* If F=MSB/MSW > F[(b-1), (a-1)(b-1)(c-1); alpha] reject H_02 */
   simple_F_test(QUERY);
   /* If F=MSC/MSW > F[(c-1), (a-1)(b-1)(c-1); alpha] reject H_03 */
   simple_F_test(RECALL);
#undef simple_F_test
   

#define complex_F_test(FIRST,SECOND) {                                       \
      const        interaction[] = {PARQUE, PARREC, QUEREC};                 \
      const int    INTERACT = interaction[FIRST + SECOND - 1];               \
      const char   before[] = "   F-test of interaction between ";           \
      const char   after[]  = " data ";                                      \
      const char   middle[] = " and ";                                       \
      const size_t baselen  = strlen(before) + strlen(after) +               \
                              strlen(middle);                                \
      const int    len      = strlen(info[FIRST].smallname) +                \
                              strlen(info[SECOND].smallname) + baselen;      \
                                                                             \
   if (showall) {                                                            \
      fputs(" ", stdout);                                                    \
      PrintLine(stdout, '-', len, "");                                       \
      NewLine(1, stdout);                                                    \
      printf("%s%s%s%s%s",                                                   \
       before, info[FIRST].smallname, middle, info[SECOND].smallname, after);\
      NewLine(1, stdout);                                                    \
      fputs(" ", stdout);                                                    \
      PrintLine(stdout, '-', len, "");                                       \
      NewLine(1, stdout);                                                    \
   }                                                                         \
   info[INTERACT].f_value = info[INTERACT].meansquare /                      \
       info[WITHIN].meansquare;                                              \
   if (showall) {                                                            \
      printf("   F = MS%c%c/MSW = "REAL_SPEC,                                \
             ('A'+FIRST), ('A'+SECOND), info[INTERACT].f_value);             \
      NewLine(1, stdout);                                                    \
   }                                                                         \
   info[INTERACT].signifdiff = F_is_significant(info[INTERACT].degfreed,     \
                                                info[WITHIN].degfreed,       \
                                                info[INTERACT].f_value,      \
                                                &(info[INTERACT].critval),   \
                                                showall);                    \
   if (showall) {                                                            \
      if (info[INTERACT].signifdiff) {                                       \
         printf("   At least one %s and %s level interaction is ",           \
                info[FIRST].smallname, info[SECOND].smallname);              \
         NewLine(1, stdout);                                                 \
         fputs ("   different from the others.", stdout);                    \
      } else {                                                               \
         fputs ("    No significant difference detected in interactions ",   \
               stdout);                                                      \
         NewLine(1, stdout);                                                 \
         printf("    between %s and %s level means.",                        \
                info[FIRST].smallname, info[SECOND].smallname);              \
      }                                                                      \
      NewLine(2, stdout);                                                    \
   }                                                                         \
} /* #complex_F_test */


   /* If F=MSAB/MSW > F[(a-1)(b-1), (a-1)(b-1)(c-1); alpha] reject H_04 */
   complex_F_test(PARTI,QUERY);
   /* If F=MSAC/MSW > F[(a-1)(c-1), (a-1)(b-1)(c-1); alpha] reject H_05 */
   complex_F_test(PARTI,RECALL);
   /* If F=MSBC/MSW > F[(b-1)(c-1), (a-1)(b-1)(c-1); alpha] reject H_06 */
   complex_F_test(QUERY,RECALL);
#undef complex_F_test

   no_signif_diff = (! (info[PARQUE].signifdiff || info[PARREC].signifdiff));

   if ((info[PARTI].signifdiff) && (! no_signif_diff)) {
      fputs("\nAt least one pair of the participant means differ.  But: ", \
            stdout);
      NewLine(1, stdout);
      if ((info[PARQUE].signifdiff) && (info[PARREC].signifdiff)) {
         fputs("significant interaction effects were detected between ", \
               stdout);
         NewLine(1, stdout);
         fputs("pairs of participant means and both query and recall means.", \
               stdout);
         NewLine(2, stdout);
      } else {
         fputs("  a significant interaction effect was detected between ", \
               stdout);
         NewLine(1, stdout);
         fputs("  pairs of participant means and ", stdout);
         if (info[PARQUE].signifdiff) {
            fputs("query means.", stdout);
         } else 
         if (info[PARREC].signifdiff) {
            fputs("recall means.", stdout);
         } else {
            fputs("...", stdout);
            NewLine(1, stdout);
            fputs("we are experiencing technical difficulty", stdout);
            no_signif_diff = TRUE;
            swerr(proc, "significant difference indicated but none found", 0);
         }
      }
      NewLine(1, stdout);
      if (! no_signif_diff) {
         printf("The participant data should be tested at fixed " \
              "levels of the other factor%s.",                    \
              (info[PARQUE].signifdiff && info[PARREC].signifdiff)?"s":"");
         NewLine(1, stdout);
      }
   }
   if (always_show_means || no_signif_diff) {
      if (showall) {
         NewLine(1, stdout);
      }
      Q = HSD(info[PARTI].signifdiff, info[PARTI].degfreed,  \
              info[QUERY].degfreed, info[WITHIN].meansquare, \
              alpha_level, showall);
      if (showall) {
         NewLine(2, stdout);
      }
      if (make_means(averages, partiname, partimean, maxparti)) {
         mean_groups(Q, averages, maxparti);
         show_means(averages, maxparti, datatype,      \
                    (transstring[0])?transstring:NULL, \
                    "participant", level_list, showall, stdout);
         NewLine(2, stdout);
      } else {
         swerr(proc, "unable to make mean groups", 0);
      }
   } /* no_significant_diff */

   ANOVATable(info, (TOTAL + 1U),                             \
              "ANOVA Table for a Difference in Mean Scores, " \
              "Three Factor Design", datatype, level_list,    \
              (transstring[0])?transstring:NULL, stdout);
   end_display();
   if (showall) { 
      if (1 < maxparti) {
         print_name_list(partiname, maxparti, 5, width_display(), \
                         "Participants: ", stdout);
      }
      if (1 < maxquery) {
         print_name_list(queryname, maxquery, 5, width_display(), \
                         "Queries: ", stdout);
      }
   }
   {
      register matrix_dim p, q;

      for (p=0; p < maxparti; p++) {
         free(parquesum[p]);
         free(parquemean[p]);
         free(parrecsum[p]);
         free(parrecmean[p]);
      }
      for (q=0; q < maxquery; q++) {
         free(querecsum[q]);
         free(querecmean[q]);
      }
   }   
   end_display();
}/* ThreeWayStats() */


/*                                                                         *\
   -----------------------------------------------------------------------
                      Compute Averages and Sum of All
   -----------------------------------------------------------------------
\*                                                                         */

void aver_compute (const bool       byparti, /* F: by query */
                         Matrix *   data[],  /* contains only selected pts */
                   const matrix_dim depth,   /* size of data[] */
                         real       aver[])  /* will hold average values */
{
            real (* sumfunc) 
                    (Matrix * arr[], const matrix_dim, const matrix_dim);
            real       divisor;
   register matrix_dim i, max;

   if (byparti) {
     sumfunc = sum3Dcol;
     divisor = num_rows(data[0]);
     max     = num_cols(data[0]);
   } else {
     sumfunc = sum3Drow;
     divisor = num_cols(data[0]);
     max     = num_rows(data[0]);
   }
   divisor *= (real)depth;

   for (i=0; i < max; i++) {
      aver[i] = (sumfunc(data, i, depth)) / (real)divisor;
   } 
} /* aver_compute() */


real sum_of_all(      Matrix *   data[],    /* contains only selected pts */
                const matrix_dim max_depth) /* size of data[] */
{
   register matrix_dim col;
   register real       sum     = 0.0;
            matrix_dim max_col = num_cols(data[0]);

   for (col=0; col < max_col; col++) {
      sum += sum3Dcol(data, col, max_depth);
   }
   return(sum);
 } /* sum_of_all() */


real aver_by_both(      Matrix *   data[],    /* contains only selected pts */
                  const matrix_dim max_depth) /* size of data[] */
{
   real divisor = max_depth * num_cols(data[0]) * num_rows(data[0]);

   return (sum_of_all(data, max_depth) / divisor);
} /* aver_by_both() */


/*                                                                         *\
   -----------------------------------------------------------------------
			       Miscellaneous
   -----------------------------------------------------------------------
\*                                                                         */

real sum(const real       vector[],
         const matrix_dim size) 
{
   register matrix_dim i;
   register real       total = 0.0;

   for (i=0; i < size; i++) {
      total += vector[i];
   }
   return (total);
} /* sum() */


real average(const real       vector[],
             const matrix_dim size) 
{
   return (sum(vector, size) / size);
} /* average() */


void extract_column(const real *     vector[],
                    const matrix_dim max_row,
                    const matrix_dim column,
                          real       chosen[]) 
{
   register matrix_dim row;

   for (row=0; row < max_row; row++) {
      chosen[row] = vector[column][row];
   }
} /* extract_column() */

/*                                                                         *\
   -----------------------------------------------------------------------

   -----------------------------------------------------------------------
\*                                                                         */

