#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#ifndef UWOCSD
double log1p(double x);
#endif
#include "types.h"
#include "proto.h"
#include "select.h"
#include "print.h"  /* for start_menu() */
#include "transform.h"

/* local definitions */
typedef struct {
                string          name;
                TransFuncType * transfunc;
               } Transform;

/* local prototypes */
static void arcsine_transform(real * X_hj);
static void logit_transform(real * X_hj);
static void nullify_transform(void);

/* module specific global variables */
#define ACTUAL FALSE
#define SAVED  TRUE

static       unsigned short which_transform[] = {1,1};
static       when_type      when              = WHEN_TEST;
static const Transform      transform_list[]  = {
                              {"no",      NULL},
                              {"arcsine", arcsine_transform},
                              {"logit",   logit_transform}};
#define NULL_TRANSFORM (0)
static const unsigned short NUM_OF_TRANSFORMS = \
      (sizeof(transform_list) / sizeof(transform_list[0]));
static const string when_str[] = {"never",
                                  "if variances differ",
                                  "before analysis begins"};

/* local prototypes */
static void tran_swap(void);

/*                                                                         *\
   -----------------------------------------------------------------------
				Transforms
   -----------------------------------------------------------------------
\*                                                                         */

void arcsine_transform(real * X_hj) 
{
   *X_hj = (real)asin( (double)sqrt( (double)(* X_hj) ) ); 
} /* arcsine_transform() */


void logit_transform(real * X_hj) 
{
   const double value  = (1.0 == *X_hj) ?(double)*X_hj - DBL_EPSILON  \
                                        :((0.0 == *X_hj) ?DBL_EPSILON \
                                                        :(double)*X_hj);
/*
   logit transform:
      log(p / 1-p)
      ranges from 0 to +oo
      when p is 1 it should be replaced by a large value (not HUGE)

   log(p / (1-p)) = log(p) - log(1-p)
   since p is already < 1 we use log1p()
   log1p(x) is log(1 + x) accurate even for small x
*/
   *X_hj = (real)(log1p(value - 1.0) - log1p(-value));
} /* log_transform() */


TransFuncType * transform(when_type current_step) 
{
   extern when_type when;

   return ((current_step == when) \
                ?transform_list[which_transform[ACTUAL]].transfunc \
		:NULL);
} /* transform() */


/*                                                                         *\
   -----------------------------------------------------------------------
                         Access to struct Elements
   -----------------------------------------------------------------------
\*                                                                         */

string when_descrip(void)
{
   extern when_type when;

   return(when_str[when]);
} /* when_descrip() */


bool never_transform(void) 
{
   extern when_type when;

   return (WHEN_NEVER == when);
} /* never_transform() */


bool always_transform(void) 
{
   extern when_type when;

   return (WHEN_ALWAYS == when);
} /* always_transform() */


string tran_name(void) 
{
   return (transform_list[which_transform[never_transform()]].name);
} /* tran_name() */


bool tran_set(const unsigned short new_transform) 
{
   extern unsigned short which_transform[];

   if (new_transform < NUM_OF_TRANSFORMS) {
      which_transform[never_transform()] = new_transform; 
      return (TRUE);
   } else {
      return (FALSE);
   }
} /* tran_set() */


void nullify_transform(void)
{
   extern unsigned short which_transform[];

   which_transform[ACTUAL] = NULL_TRANSFORM;
} /* nullify_transofrm() */


void tran_swap(void) 
{
   extern unsigned short which_transform[];

   if (NULL_TRANSFORM != which_transform[ACTUAL]) {
      which_transform[SAVED] = which_transform[ACTUAL];
      nullify_transform();
   } else {
      which_transform[ACTUAL] = which_transform[SAVED];
   }
} /* tran_swap() */


/*                                                                         *\
   -----------------------------------------------------------------------
                              Transform Menus
   -----------------------------------------------------------------------
\*                                                                         */

void tran_choose(const char   product[],
                 const char   version[],
                 const string menutitle)
{
            const char           proc[] = "tran_choose";
   extern   const Transform      transform_list[];
   extern   const unsigned short NUM_OF_TRANSFORMS;
   extern         unsigned short which_transform[];
   extern         char           PROMPT[];
   extern         char           errorstring[];
   register       unsigned short i;
                  unsigned short selected_transform;
                  unsigned short choice;
                  bool           which = never_transform();

   start_menu();
   printf("%s -- version %s\n\n", product, version);
   puts(menutitle);
   puts("Select transform to use");
   for (i=1; i < NUM_OF_TRANSFORMS; i++) {
      printf("   %c %u. %s\n",
             (i == which_transform[which])?on:off, i, transform_list[i].name);
      if (i == which_transform[which]) {
         selected_transform = i;
      }
   }
   choice = getchoice(1, NUM_OF_TRANSFORMS - 1U, PROMPT, selected_transform);
      if (0 != choice) {
         if (! tran_set(choice)) {
            sprintf(errorstring, "error changing to transform #%hu!", choice);
            swerr(proc, errorstring, 1);
            /* NOTREACHED */
	 }
      } else {
         swerr(proc, "error from getchoice(), transform not changed", 0);
      }
} /* tran_choose() */


void when_trans_menu(const char   product[],
                     const char   version[],
                     const string menutitle)
{
            const char           proc[] = "when_trans_choose";
   extern         when_type      when;
   extern         char           PROMPT[];
   static   const unsigned short WHEN_MAX = WHEN_ALWAYS + 1U;
                  unsigned short choice;
                  unsigned short i;

   start_menu();
   printf("%s -- version %s\n\n", product, version);
   puts(menutitle);
   puts("Select when to use transform");
   for (i=0; i < WHEN_MAX; i++) {
      printf("   %c %u. %s\n",
             (i == when)?on:off, i+1, when_str[i]);
   }
   choice = getchoice(1, (elem_t)WHEN_MAX, PROMPT, when);
      if (0 != choice) {
         choice--;
         if ((never_transform() && (WHEN_NEVER != choice)) || \
            ((! never_transform()) && (WHEN_NEVER == choice))) {
            tran_swap();
	 }
         when = choice;
      } else {
         swerr(proc, "error from getchoice(), `when' not changed", 0);
      }
} /* when_trans_choose() */


/*                                                                         *\
   -----------------------------------------------------------------------

   -----------------------------------------------------------------------
\*                                                                         */

